/////////////////////////////////////////////////////////////////////////////
// Copyright (C) <%year%> The Linux Foundation. All rights reserved.
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// This file contains tests for the following library: 
// <%library%>, 
// section: "<%libsection%>".
// 
// This C file was generated by the T2C system developed in ISPRAS 
// for The Linux Foundation. 
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// To build a standalone version of the tests, i.e. the one that does not
// depend on TETware Lite test harness, please follow the instructions below.
// (A standalone executable like this can prove useful for debugging.) 
// 
// 1. Make sure the T2C_ROOT environment variable is set properly. It should 
// contain a path to the main test suite directory. For instance, the T2C 
// code generator executable (t2c) resides in "$T2C_ROOT/tools/t2c/bin", 
// the tet execution configuration file (tetexec.cfg) can be found right in
// "$T2C_ROOT/" as well as "tet_code" file etc.
// 
// 2. The standalone version of the tests will be built with lsbcc, so make sure 
// that lsbcc is somewhere in your PATH (it usually resides in /opt/lsb/bin). 
// You should also add a path to the LSB pkg-config data 
// (usually /opt/lsb/lib/pkgconfig or /opt/lsb/lib64/pkgconfig) to the beginning of 
// the string contained in PKG_CONFIG_PATH environment variable.
// 
// 3. Execute the following commands:
//      cd $T2C_ROOT/<%suite_subdir%>/tests/<%object_name%>/
//      make clean
//      make debug
// 
// The C-code of the tests is the same both for release and debug (standalone)
// versions.
// main() function for standalone execution is defined in
// "$T2C_ROOT/tools/t2c/debug/src/dbg_main.c" along with some utility stuff.
// The TET API stubs are declared in "$T2C_ROOT/tools/t2c/debug/include/tet_api.h"
/////////////////////////////////////////////////////////////////////////////

#include <errno.h>
#include <unistd.h>

#include <t2c.h>

// global variables
int nPurposesTotal = 0;      // test purpose count
int nPurposesPassed = 0;     // number of test purposes that passed
unsigned test_passed_flag = TRUE;

int bVerbose = FALSE;      

const char* test_name_      = "<%object_name%>";
const char* suite_subdir_   = "<%suite_subdir%>";

const char* rel_href_path_  = 
    "<%suite_subdir%>/tests/<%object_name%>/<%object_name%>.html#%s%d\">";
char* t2c_href_tpl_ = NULL;
char* t2c_href_full_tpl_ = NULL;

// Name of the env. variable that is used to determine whether the REQ
// implementation should output just IDs (value = 0, default) or hyperlinks
// to the html representation, enclosed in some pseudotags for DTK Manager
// (value != 0).
const char* t2c_gen_hlinks_name = "T2C_GEN_HLINKS";
int gen_hlinks = 0;

TReqInfoList* head_ = NULL;  // A list of REQs. Use it to load and free REQs.
TReqInfoPtr*  reqs_ = NULL;  // An array of pointers to the TReqInfo structures.
int nreq_ = 0;               // Total number of REQs (length of the array).


// function prototypes
static void test_passed();
static void startup_func();
static void cleanup_func();

static void user_startup();
static void user_cleanup();

static void tp_launcher();

// other globals 
<%globals%>

// User-defined startup instructions
static void
user_startup(int* us_failed, char** reason)
{
    int init_failed = FALSE;
    char* reason_to_cancel = NULL;
    
<%startup%>
    if (init_failed) *us_failed = TRUE;
    if (reason_to_cancel) *reason = reason_to_cancel;
}

// User-defined cleanup instructions
static void
user_cleanup()
{
<%cleanup%>
}
 
///////////////////////////////////////////////////////////////////////////
// Test purposes
///////////////////////////////////////////////////////////////////////////
 
<%test_purposes%>

///////////////////////////////////////////////////////////////////////////
// Specials
///////////////////////////////////////////////////////////////////////////
// This function is called for each test that passes.
static void 
test_passed()
{
    if (test_passed_flag)
    {
        nPurposesPassed++;
    }
    tet_result(TET_PASS);
}

struct tet_testlist tet_testlist[] = {
<%tet_hooks%>    {NULL, 0}
};

// The list of tests in this file. 
TTestPurposeType test_purpose_func[] = {
<%tp_funcs%>    NULL
};
    
// Parent control functions.
TParentControlFunc pc_func[] = {
<%pcf_funcs%>    NULL
};

// Test purpose launcher.
static void 
tp_launcher()
{
    int tp_ind = tet_thistest - 1;
    int result = t2c_fork(
        test_purpose_func[tp_ind],  // a test purpose to launch
        pc_func[tp_ind],            // parent control func
        <%wait_time%>, // wait time (seconds)
        NULL, NULL);
        //user_startup, user_cleanup);
    
    if (result == -1)
    {
        tet_result(TET_UNRESOLVED);
        test_passed_flag = FALSE;
    }
    else
    {
        test_passed_flag = result;
    }
    test_passed();
}

// TET startup function
static void 
startup_func()
{
    char* tsf_verb_ = NULL;
    char* reason_to_cancel = NULL;
    
    int init_failed = FALSE;
    
    // Determine the value of the VERBOSE variable (if present). 
    tsf_verb_ = tet_getvar("VERBOSE");
    
    if ((tsf_verb_ != NULL) && !strcmp(tsf_verb_, "yes"))
    {
        bVerbose = TRUE;
    }
    
    // Get a path to the REQ catalogue.
    char* rcat_path_ = t2c_get_rcat_path(suite_subdir_, test_name_);
    int bOK = t2c_rcat_load(rcat_path_, &head_, &reqs_, &nreq_);
    free(rcat_path_);
    
    if (!t2c_root_defined())
    {
        INIT_FAILED("T2C_ROOT environment variable is not defined.");
    }
    else
    {
        char* t2c_root = t2c_get_root();
        char* beg = "{noreplace}<a href=\"";
        char* stmp = NULL;
        
        stmp = str_sum(beg, t2c_root);
        t2c_href_tpl_ = str_sum(stmp, rel_href_path_);
        
        t2c_href_full_tpl_ = str_sum(t2c_href_tpl_, "%s</a>{/noreplace}");
        
        free(t2c_root);
        free(stmp);
    }
    
#ifndef T2C_IGNORE_RCAT_ERRORS
    if (!bOK)
    {
        INIT_FAILED("<%object_name%>: unable to load req catalogue.");
    }
#endif
    
    // Sort the catalogue by ID. (No-op if reqs == NULL.)
    t2c_rcat_sort(reqs_, nreq_);
    
    int pcc_res = pipe(pcc_pipe_);
    if (pcc_res != 0)
    {
        INIT_FAILED("Unable to create parent-child control pipe.");
    }
    
    char* glh_tmp = getenv(t2c_gen_hlinks_name);
    if (!glh_tmp)
    {
        gen_hlinks = 0;
    }
    else
    {
        gen_hlinks = atoi(glh_tmp);
    }
    
    // Perform user-defined startup instructions.
    user_startup(&init_failed, &reason_to_cancel);
    
    if (init_failed)
    {
        // Test case initialization failed. Cancel all test purposes.
        int num_purp_ = sizeof(tet_testlist) / sizeof(tet_testlist[0]) - 1;
        int cur_purp_ = 1;
        
        for (; cur_purp_ <= num_purp_; ++cur_purp_)
        {
            tet_delete(cur_purp_, reason_to_cancel);
        }
    }
    
    fprintf(stderr, "\nExecuting tests for %s\n", test_name_);
}
 
// TET cleanup function
static void 
cleanup_func()
{
    // Perform user-defined cleanup instructions.
    user_cleanup();
    
    int num_purp_ = sizeof(tet_testlist) / sizeof(tet_testlist[0]) - 1;
    
    fprintf(stderr, "\nPassed %d of %d\n", nPurposesPassed, num_purp_);
    t2c_req_info_list_clear(head_);
    free(reqs_);
    
    free(t2c_href_tpl_);    
    free(t2c_href_full_tpl_);
    
    close(pcc_pipe_[0]);
    close(pcc_pipe_[1]);
}

// Tell TET to use startup & cleanup functions we provide.
void (*tet_startup)() = startup_func;
void (*tet_cleanup)() = cleanup_func;
